-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local Crafting = TSM.Operations:NewPackage("Crafting")
local Operation = TSM.LibTSMTypes:Include("Operation")
local Group = TSM.LibTSMTypes:Include("Group")
local CustomString = TSM.LibTSMTypes:Include("CustomString")
local L = TSM.Include("Locale").GetTable()
local ChatMessage = TSM.Include("Util.ChatMessage")
local CustomPrice = TSM.Include("Service.CustomPrice")
local ItemInfo = TSM.LibTSMService:Include("Item.ItemInfo")
local private = {}
local MIN_RESTOCK_VALUE = 0
local MAX_RESTOCK_VALUE = 2000
local BAD_CRAFTING_PRICE_SOURCES = {
	crafting = true,
}



-- ============================================================================
-- Module Functions
-- ============================================================================

function Crafting.OnInitialize()
	local operationType = Operation.NewType("Crafting", L["Crafting"], 1)
		:SetDescriptionFunc(private.GetDescription)
		:AddStringSetting("minRestock", "10")
		:AddStringSetting("maxRestock", "20")
		:AddStringSetting("minProfit", "100g")
		:AddStringSetting("craftPriceMethod", "")
	TSM.Operations.RegisterType(operationType)

	for _, name in Operation.Iterator("Crafting") do
		local operation = Operation.GetSettings("Crafting", name)
		if operation.craftPriceMethod ~= "" then
			local isValid, err = CustomPrice.Validate(operation.craftPriceMethod, BAD_CRAFTING_PRICE_SOURCES)
			if not isValid then
				ChatMessage.PrintfUser(L["Your craft value method for '%s' was invalid so it has been returned to the default. Details: %s"], name, err)
				operation.craftPriceMethod = ""
			end
		end
	end
end

function Crafting.HasOperation(itemString)
	return private.GetOperationSettings(itemString) and true or false
end

function Crafting.GetRestockRange()
	return MIN_RESTOCK_VALUE, MAX_RESTOCK_VALUE
end

function Crafting.IsValid(itemString)
	local origItemString = itemString
	itemString = Group.TranslateItemString(itemString)
	local operationName, operationSettings = TSM.Operations.GetFirstOperationByItem("Crafting", itemString)
	if not operationSettings then
		return false
	end
	local minRestock, maxRestock, errMsg = nil, nil, nil
	minRestock, errMsg = private.GetMinRestock(operationSettings, origItemString)
	if not minRestock then
		return false, errMsg
	end
	maxRestock, errMsg = private.GetMaxRestock(operationSettings, origItemString)
	if not maxRestock then
		return false, errMsg
	end
	if minRestock > maxRestock then
		-- invalid cause min > max restock quantity
		return false, format(L["'%s' is an invalid operation. Min restock of %d is higher than max restock of %d for %s."], operationName, minRestock, maxRestock, ItemInfo.GetLink(origItemString))
	end
	return true
end

function Crafting.GetMinProfit(itemString)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return false
	end
	if operationSettings.minProfit == "" then
		return false
	end
	local value = CustomString.GetValue(operationSettings.minProfit, itemString)
	return true, value
end

function Crafting.GetRestockQuantity(itemString, haveQuantity)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return 0
	end
	local minRestock = private.GetMinRestock(operationSettings, itemString)
	local maxRestock = private.GetMaxRestock(operationSettings, itemString)
	if not minRestock or not maxRestock or minRestock > maxRestock then
		return 0
	end
	local neededQuantity = maxRestock - haveQuantity
	if neededQuantity <= 0 then
		-- don't need to queue any
		return 0
	elseif neededQuantity < minRestock then
		-- we're below the min restock quantity
		return 0
	end
	return neededQuantity
end

function Crafting.GetCraftedItemValue(itemString)
	local operationSettings = private.GetOperationSettings(itemString)
	if not operationSettings then
		return false
	end
	if operationSettings.craftPriceMethod == "" then
		return false
	end
	local value = CustomString.GetValue(operationSettings.craftPriceMethod, itemString)
	return true, value
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.GetMinRestock(operationSettings, itemString)
	local minRestock, errStr = CustomPrice.GetValue(operationSettings.minRestock, itemString, true)
	if not minRestock then
		return nil, format(L["Your min restock (%s) is invalid for %s."], operationSettings.minRestock, ItemInfo.GetLink(itemString) or "?").." "..errStr
	elseif minRestock < MIN_RESTOCK_VALUE or minRestock > MAX_RESTOCK_VALUE then
		return nil, format(L["Your min restock (%s) is invalid for %s."], operationSettings.minRestock, ItemInfo.GetLink(itemString) or "?").." "..format(L["Must be between %d and %s."], MIN_RESTOCK_VALUE, MAX_RESTOCK_VALUE)
	end
	return minRestock
end

function private.GetMaxRestock(operationSettings, itemString)
	local maxRestock, errStr = CustomPrice.GetValue(operationSettings.maxRestock, itemString, true)
	if not maxRestock then
		return nil, format(L["Your max restock (%s) is invalid for %s."], operationSettings.maxRestock, ItemInfo.GetLink(itemString) or "?").." "..errStr
	elseif maxRestock < MIN_RESTOCK_VALUE or maxRestock > MAX_RESTOCK_VALUE then
		return nil, format(L["Your max restock (%s) is invalid for %s."], operationSettings.maxRestock, ItemInfo.GetLink(itemString) or "?").." "..format(L["Must be between %d and %s."], MIN_RESTOCK_VALUE, MAX_RESTOCK_VALUE)
	end
	return maxRestock
end

function private.GetDescription(operationSettings)
	if operationSettings.minProfit ~= "" then
		return L["Restocking with a min profit."]
	else
		return L["Restocking with no min profit."]
	end
end

function private.GetOperationSettings(itemString)
	itemString = Group.TranslateItemString(itemString)
	local operationName, operationSettings = TSM.Operations.GetFirstOperationByItem("Crafting", itemString)
	if not operationName then
		return
	end
	return operationSettings
end
